﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Dynamic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Xml;

namespace DynamicToolkit
{
    /// <summary>
    /// This class allows you to directly navigate XML documents in a similar fashion to XPath using .NET.
    /// </summary>
    /// <example>
    ///     // If you intend to use lambda expressions, you'll need to declare an IPerson interface to 
    ///     // work around some C# limitations.
    ///                 
    ///     // Resolves to: "/*/Contacts/Contact[Name = 'Tony']/Phone[1]/text()"
    ///     string tonyPhoneNumber = addressBook.Contacts.Contact["Name = 'Tony'"].Phone;
    ///     string tonyPhoneNumber2 = addressBook.Contacts.Contact["Name = 'Tony'"].Phone.Value();
    ///     string tonyPhoneNumber3 = addressBook.Contacts.Contact["Name = 'Tony'"].Phone[0].Value();
    ///
    ///     // Resolves to: "/*/Contacts/Contact[Name = 'Steph']/Phone[1]/text()"
    ///     // Currently not in supported in C# 4.0
    ///     // C# 4.0 does not support operations on dynamic types in Expression&lt;TDelegate&gt;
    ///     //string stephPhoneNumber = addressBook.Contacts.Contact[() =&gt; p.Name == 'Steph'].Phone;
    ///     //string stephPhoneNumber2 = addressBook.Contacts.Contact[p =&gt; p.Name == 'Steph'].Phone.Value();
    ///
    ///     // Resolves to: "/*/Contacts/Contact[Name = 'Steph']/Phone[1]/text()"
    ///     // Workaround for the above C# 4.0 limitation using IPerson interface and declaring Expression outside of dynamic object
    ///     Expression&lt;Func&lt;IPerson, bool&gt;&gt; nameEqualsSteph = (p) =&gt; p.Name == "Steph";
    ///     string stephPhoneNumber3 = addressBook.Contacts.Contact[nameEqualsSteph].Phone;
    ///
    ///     // Resolves to: "/*/Contacts/Contact[Name = 'Steph']/Phone[1]/text()"
    ///     // Currently not in supported in C# 4.0
    ///     //List&lt;string&gt; tonyAndStephsPhoneNumbers = addressBook.Contacts.Contact[p =&gt; p.Name == "Tony" || p.Name == 'Steph'].Phone;
    ///
    ///     // Resolves to: "/*/Contacts/Contact[Name = 'Tony' or Name = 'Steph']/Phone[1]/text()"
    ///     // Workaround for the above C# 4.0 limitation.  We use a helper interface (IPerson) to get around some of the limitations.
    ///     // String comparisions (greater than, less than, ...) are supported via the DynamicTools.ComparableString class.
    ///     // There is speculation that C# will support extension operator overloading in a future version.
    ///     Expression&lt;Func&lt;IPerson, bool&gt;&gt; nameEqualsTonyOrSteph = (p) =&gt; p.Name == "Tony" || p.Name == "Steph";
    ///     List&lt;string&gt; tonyAndStephsPhoneNumbers2 = addressBook.Contacts.Contact[nameEqualsTonyOrSteph].Phone;
    ///
    ///     // Resolves to: "/*/Contacts/Contact/Phone"
    ///     XmlNodeList allPhoneNumberNodes = addressBook.Contacts.Contact.Phone;
    ///
    ///     // Resolves to: "/*/Contacts/Contact/Phone/text()"
    ///     List&lt;string&gt; allPhoneNumbers = addressBook.Contacts.Contact.Phone;
    ///     IEnumerable&lt;string&gt; allPhoneNumbers2 = addressBook.Contacts.Contact.Phone.Values();
    ///
    ///     // Resolves to: "/*/Contacts/Contact[1]"
    ///     XmlNode firstContact = addressBook.Contacts.Contact;
    ///     XmlNode firstContact2 = addressBook.Contacts.Contact.Node();
    ///     XmlNode firstContact3 = addressBook.Contacts.Contact[0].Node();
    ///
    ///     // Resolves to: "/*/Contacts/Contact[Name = 'Tony']/Phone[1]/@Tag"
    ///     string phoneTag = addressBook.Contacts.Contact["Name = 'Tony'"].Phone.Attribute("Tag");
    ///
    ///     // Enumerating over the dynamic XPath
    ///     foreach (dynamic contact in addressBook.Contacts.Contact)
    ///     {
    ///         Console.WriteLine(contact.Name);
    ///     }
    /// </example>
    public sealed class DynamicXpath : DynamicObject, IEnumerable<DynamicXpath>
    {
        private readonly XmlDocument doc;
        private readonly DynamicXpath parent;
        private readonly string contextStep;
        private WeakReference results;


        /// <summary>
        /// Construct a dynamic XPath based on the specified XML document
        /// </summary>
        /// <param name="doc">The XML document to query</param>
        private DynamicXpath(XmlDocument doc)
            : this(null, Step.Root(doc))
        {
            if (doc == null)
            {
                throw new ArgumentNullException("doc", "Document cannot be null");
            }

            this.doc = doc;
        }

        /// <summary>
        /// Construct a dynamic XPath based on the specified XML element
        /// </summary>
        /// <param name="element">The XML element to start query from</param>
        private DynamicXpath(XmlElement element)
            : this(null, Step.Root(element))
        {
            if (element == null)
            {
                throw new ArgumentNullException("element", "Element cannot be null");
            }

            this.doc = element.OwnerDocument;
        }

        /// <summary>
        /// Construct a child dynamic XPath belonging to the specified parent.
        /// </summary>
        /// <param name="parent">The parent dyanmic XPath</param>
        /// <param name="step">The XPath step</param>
        private DynamicXpath(DynamicXpath parent, string step)
        {
            this.parent = parent;
            this.contextStep = step;

            if (parent != null)
            {
                this.doc = parent.doc;
            }
        }

        /// <summary>
        /// Creates a child selector.  The index is zero-based.
        /// </summary>
        /// <param name="index">The index of the child</param>
        /// <returns>The resulting dynamic XPath</returns>
        public DynamicXpath this[int index]
        {
            get
            {
                return new DynamicXpath(this, Step.Index(index));
            }
        }

        /// <summary>
        /// Create a predicate
        /// </summary>
        /// <param name="predicate">The predicate string</param>
        /// <returns>The resulting dynamic XPath</returns>
        public DynamicXpath this[string predicate]
        {
            get
            {
                return new DynamicXpath(this, Step.Predicate(predicate));
            }
        }

        /// <summary>
        /// Creates an XPath predicate based on the specified predicate expression.
        /// Currently not supported by C# 4.0.
        /// </summary>
        /// <remarks>
        /// Maybe dynamic types will be supported in labda expression in future C# release.
        /// </remarks>
        /// <param name="predicate">A binary Expression</param>
        /// <returns>The resulting dynamic XPath</returns>
        public DynamicXpath this[LambdaExpression predicate]
        {
            get
            {
                return new DynamicXpath(this, Step.Predicate(predicate));
            }
        }

        /// <summary>
        /// Creates an XPath predicate based on the specified example objects.
        /// </summary>
        /// <param name="exampleObjects">Example objects to match</param>
        /// <returns>The resulting dynamic XPath</returns>
        public DynamicXpath this[params object[] exampleObjects]
        {
            get
            {
                Example[] examples = exampleObjects.Select(e => new Example(e)).ToArray();
                return this[examples];
            }
        }

        /// <summary>
        /// Creates an XPath predicate based on the specified example.
        /// </summary>
        /// <param name="examples">Example to match</param>
        /// <returns>The resulting dynamic XPath</returns>
        public DynamicXpath this[params Example[] examples]
        {
            get
            {
                return new DynamicXpath(this, Step.Predicate(examples));
            }
        }

        /// <summary>
        /// Create a dyanic XPath to query the specified document
        /// </summary>
        /// <param name="doc">The document to create DynamicXPath from</param>
        /// <returns>A dynamic XPath to query the specified document</returns>
        public static dynamic Load(XmlDocument doc)
        {
            return new DynamicXpath(doc);
        }

        /// <summary>
        /// Create a dyanic XPath to query the specified element
        /// </summary>
        /// <param name="element">The element to create DynamicXPath from</param>
        /// <returns>A dynamic XPath to query the specified element</returns>
        public static dynamic Load(XmlElement element)
        {
            return new DynamicXpath(element);
        }
        
        /// <summary>
        /// Create a dyanic XPath to query the specified file
        /// </summary>
        /// <param name="filename">The file to query</param>
        /// <returns>A dynamic XPath to query the specified file</returns>
        public static dynamic Load(string filename)
        {
            XmlDocument doc = new XmlDocument();
            doc.Load(filename);
            return new DynamicXpath(doc);
        }

        /// <summary>
        /// Create a dyanic XPath to query the specified xml data
        /// </summary>
        /// <param name="xml">The xml data query</param>
        /// <returns>A dynamic XPath to query the specified xml data</returns>
        public static dynamic LoadXml(string xml)
        {
            XmlDocument doc = new XmlDocument();
            doc.LoadXml(xml);
            return new DynamicXpath(doc);
        }

        /// <summary>
        /// Get the document
        /// </summary>
        /// <returns></returns>
        public XmlDocument GetDocument()
        {
            return this.doc;
        }

        /// <summary>
        /// Get the root dyanamic XPath
        /// </summary>
        /// <returns>A dynamic XPath that starts at the root of the document</returns>
        public DynamicXpath GetRoot()
        {
            return new DynamicXpath(this.doc);
        }

        /// <summary>
        /// Used internally.  Do not call this method.
        /// </summary>
        public override bool TryGetMember(GetMemberBinder binder, out object result)
        {
            result = new DynamicXpath(this, Step.Child(binder.Name));
            return true;
        }


        /// <summary>
        /// Get the generated XPath
        /// </summary>
        /// <returns>The XPath statement represented by the dyanmic XPath</returns>
        public string ToXpath()
        {
            Stack<DynamicXpath> nodes = new Stack<DynamicXpath>();
            DynamicXpath cur = this;

            while (cur != null)
            {
                nodes.Push(cur);
                cur = cur.parent;
            }

            StringBuilder buf = new StringBuilder();

            foreach (var node in nodes)
            {
                buf.Append(node.contextStep);
            }

            return buf.ToString();
        }
        
        /// <summary>
        /// Get the nodes matched by this dynamic XPath
        /// </summary>
        /// <returns>The matching nodes</returns>
        public XmlNodeList Nodes()
        {
            XmlNodeList nodes = null;

            if (this.results != null)
            {
                nodes = this.results.Target as XmlNodeList;
            }

            if (nodes == null)
            {
                nodes = this.doc.SelectNodes(this.ToXpath());
                this.results = new WeakReference(nodes);
            }

            return nodes;
        }

        /// <summary>
        /// Get the first node matched by this dynamic XPath
        /// </summary>
        /// <returns>The first node matching this dynamic XPath</returns>
        public XmlNode Node()
        {
            XmlNodeList nodes = this.Nodes();
            return (nodes != null && nodes.Count > 0) ? nodes[0] : null;
        }

        /// <summary>
        /// Get the attributes with the given name for all matching nodes
        /// </summary>
        /// <param name="attributeName">The attribute name</param>
        /// <returns>All the attributes with the given name for all the matched nodes</returns>
        public IEnumerable<string> Attributes(string attributeName)
        {
            DynamicXpath xpath = new DynamicXpath(this, Step.Attribute(attributeName));
            XmlNodeList nodes = xpath.Nodes();

            if (nodes != null)
            {
                foreach (XmlAttribute attribute in nodes)
                {
                    if (attribute != null)
                    {
                        yield return attribute.Value;
                    }
                }
            }
        }

        /// <summary>
        /// Gets the attribute with the given name for the first node matched by this dynamic XPath
        /// </summary>
        /// <param name="attributeName">The attribute name</param>
        /// <returns>The attribute with the given name for the first node matched by this dynamic XPath</returns>
        public string Attribute(string attributeName)
        {
            XmlNode node = this.Node();

            if (node != null)
            {
                XmlAttribute attribute = node.Attributes[attributeName];

                if (attribute != null)
                {
                    return attribute.Value;
                }
            }

            return null;
        }

        /// <summary>
        /// Gets the value for the first node matched by this dynamic XPath
        /// </summary>
        /// <returns>The value for the first node matched by this dynamic XPath</returns>
        public string Value()
        {
            XmlNode node = this.Node();

            if (node != null)
            {
                return (node is XmlElement) ? node.InnerText : node.Value;
            }

            return null;
        }

        /// <summary>
        /// Gets the value in the specified type for the first node matched by this dynamic XPath
        /// </summary>
        /// <typeparam name="TResult">The return type</typeparam>
        /// <returns>The value in the specified type for the first node matched by this dynamic XPath</returns>
        public TResult Value<TResult>()
        {
            string value = this.Value();
            TResult result = default(TResult);

            if (value != null)
            {
                result = (TResult) Convert.ChangeType(value, typeof(TResult));
            }

            return result;
        }

        /// <summary>
        /// Gets the values for the nodes matched by this dynamic XPath
        /// </summary>
        /// <returns>The values for the nodes matched by this dynamic XPath</returns>
        public IEnumerable<string> Values()
        {
            foreach (XmlNode node in this.Nodes())
            {
                yield return (node is XmlElement) ? node.InnerText : node.Value;
            }
        }

        /// <summary>
        /// Gets the values in the specified type for the nodes matched by this dynamic XPath
        /// </summary>
        /// <typeparam name="TResult">The return type</typeparam>
        /// <returns>The values in the specified type for the nodes matched by this dynamic XPath</returns>
        public IEnumerable<TResult> Values<TResult>()
        {
            Type resultType = typeof(TResult);

            foreach (string data in this.Values())
            {
                TResult result = (TResult)Convert.ChangeType(data, resultType);
                yield return result;
            }
        }

        /// <summary>
        /// Get the number of nodes in the node list
        /// </summary>
        /// <returns>The number of nodes in the node list</returns>
        public int Count()
        {
            return this.Nodes().Count;
        }

        /// <summary>
        /// Implicity convert dynamic XPath to a string
        /// </summary>
        /// <param name="xpath">The dynamic xpath</param>
        /// <returns>The value of the first node matching the dynamic xpath</returns>
        public static implicit operator string(DynamicXpath xpath)
        {
            return xpath.Value();
        }

        /// <summary>
        /// Implicity convert dynamic XPath to its list of matching string values
        /// </summary>
        /// <param name="xpath">The dynamic xpath</param>
        /// <returns>The list of values for the nodes matching the dynamic xpath</returns>
        public static implicit operator List<string>(DynamicXpath xpath)
        {
            return xpath.Values().ToList();
        }

        /// <summary>
        /// Implicity convert dynamic XPath an XML node list
        /// </summary>
        /// <param name="xpath">The dynamic xpath</param>
        /// <returns>The nodes matching the dynamic xpath</returns>
        public static implicit operator XmlNodeList(DynamicXpath xpath)
        {
            return xpath.Nodes();
        }

        /// <summary>
        /// Implicity convert dynamic XPath an XML node
        /// </summary>
        /// <param name="xpath">The dynamic xpath</param>
        /// <returns>The first node matching the dynamic xpath</returns>
        public static implicit operator XmlNode(DynamicXpath xpath)
        {
            return xpath.Node();
        }


        #region IEnumerable<DynamicXpath> Members

        /// <summary>
        /// Get enumerator to enumerate over all the dynamic XPaths from xpath[0] to xpath[n - 1].
        /// </summary>
        /// <returns>An enumerator to enumerate over all the dynamic XPaths</returns>
        public IEnumerator<DynamicXpath> GetEnumerator()
        {
            XmlNodeList nodes = this.Nodes();

            if (nodes != null)
            {
                for (int i = 0; i < nodes.Count; i++ )
                {
                    yield return new DynamicXpath(this, Step.Index(i));
                }
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return this.GetEnumerator();
        }

        #endregion
    }
}